/****************************************************************************
 * Projekt - Messtechnik 1
 * (c) 2017 softwareentwicklung-als-prozess.de
 * Dokumentation s. Blog
 * Funktionalitaet:
 * Gesamtablauf in Setup() und Loop()
 * Bedienung der Tasten zur Eingabe von Datum und Zeit und die Ausgabe von 
 * Datum, Zeit und Messwerten auf dem Display des Adafruit LCD Shields
 ***************************************************************************/
// Includes -----------------------------------------------------------------
// fuer LCD Display und Tasten
#include <Wire.h>
#include <Adafruit_RGBLCDShield.h>
#include <utility/Adafruit_MCP23017.h>
// fuer alle weiteren komponenten
#include "sommerzeit.h"
#include "webserver.h"
#include "debug.h"
#include "globals.h"
#include "rtclock.h"
#include "sensordht11.h"
#include "ntpclient.h"
#include "mqttclient.h"

// Defines -----------------------------------------------------------------
// fuer backlight color
#define OFF 0x0
#define ON 0x1
#define RED 0x1
#define YELLOW 0x3
#define GREEN 0x2
#define TEAL 0x6
#define BLUE 0x4
#define VIOLET 0x5
#define WHITE 0x7
// fuer Wartezeiten
#define MINUTE 60000  // in millisekunden
// fuer Cursorposition
#define CURSORMARGINLEFT 0
#define CURSORMARGINRIGHT 13

// Variablen -----------------------------------------------------------------
// Wartezeiten
unsigned long starttime  = (unsigned long)0;
unsigned long actualtime = (unsigned long)0;
// Tastenstatus
boolean buttonpressed       = (boolean)false; // keine Taste gedrueckt
boolean selectbuttonpressed = (boolean)false; // Select Taste nicht gedrueckt
// Cursorposition
byte curpos = (byte)CURSORMARGINLEFT;
// LCD Shield mit Tasten (I2C)
Adafruit_RGBLCDShield lcd = Adafruit_RGBLCDShield();

// Funktionen ---------------------------------------------------------------
void setup() 
{
  // Debugmeldungen
  DebugPrintInit();
  DebugPrint("Projekt Messtechnik1");
  // LCD
  lcd.begin(16, 2);                              // Anzahl Spalten und Zeilen vom Display setzen
  lcd.setBacklight(ON);                          // Beleuchtung einschalten
  lcd.clear();                                   // erstmal alles loeschen
  lcd.home();                                    // Cursor nach links oben
  ActualizeDisplay();                            // Werte ausgeben
  // RTC
  RTCActivate();                                 // RTC aktivieren
  RTCCheckStatus();                              // RTC Status einlesen
  RTCRead();                                     // RTC einlesen
  ActualizeDisplay();                            // eingelesene Werte ausgeben
  SaveTimeToFallback();                          // aktuelle Zeit merken
  RTCStart();                                    // RTC Clock starten
  // DHT11
  DHT11Activate();                               // Temp-/Feuchtesensor aktivieren
  DHT11Start();                                  // Sensor starten
  DHT11GetInfoTemperature();                     // Sensorinformation einlesen
  DHT11GetInfoHumidity();                        // Sensor Messwerte einlesen
  DebugPrint("----------");
  DHT11Wait();
  Set_Temperatur(DHT11ReadTemperature());        // Temperaturwert
  Set_Luftfeuchte(DHT11ReadHumidity());          // relative Luftfeuchte
  ActualizeDisplay();                            // eingelesene Werte ausgeben
  // Ethernet / MQTT / NTP / HTTPServer
  if (MQTTSetup() == false)
  {
    DebugPrint("Ethernet Fehler - DHCP nicht erfolgreich");
  }
  else
  {
    DebugPrint("- Ethernet aktiviert -");
    DebugPrintIPAddress();
    NTP_Get();                                   // NTP Abfrage
    RTCWrite();                                  // RTC mit neuer Zeit setzen
    WEB_Init();                                  // HTTP Server starten
  }
  starttime = actualtime = millis();             // naechste Minute Wartezeit starten
  DebugPrint("*** Setup abgeschlossen ***");
}

void loop() 
{
  uint8_t buttons = (uint8_t)0;
  
  WEB_Server();                                   // Messwerte im Browser ausgeben
  if ((actualtime - starttime) > MINUTE)         // Eine Minute ist abgelaufen
  {
    DebugPrint("\n-- Minute abgelaufen: RTC und Messwerte einlesen");
    RTCRead();                                   // RTC einlesen
                                                 // Messwerte einlesen
    Set_Temperatur(DHT11ReadTemperature());      // Temperaturwert
    Set_Luftfeuchte(DHT11ReadHumidity());        // relative Luftfeuchte
    CursorBlinkAus();                            // Cursor und Blinken aus
    ActualizeDisplay();                          // Werte ausgeben
    starttime     = millis();                    // Start fuer naechste Minute
    MQTTConnectAndPublish();                     // Messwerte zum MQTT Server senden
    if (buttonpressed == true)                   // Aenderung der Zeit
    {                                            // Abfrage erfolgt eine Minute spaeter
      if (selectbuttonpressed == false)          // Aenderung wurde nicht uebernommen
      {
        RestoreFallbackToTime();                 // Alte Zeit wiederherstellen
        buttonpressed = false;                   // zuruek auf keine Aenderung
      }
    }
  }                                              // Innerhalb der minute
  actualtime = millis();                         // aktuelle Zeit einlesen
  CheckTimeOverrun();                            // Ueberlauf von millis() oder Programmfehler
  buttons = GetButtons();                        // Tasten einlesen
  CheckButtons(buttons);                         // Tasten vorauswerten
  if (selectbuttonpressed == true)               // Select Taste wurde gedrueckt
  {
    selectbuttonpressed = false;                 // Select Taste fertig ausgewertet
    RTCWrite();                                  // RTC mit neuer Zeit setzen
    CursorBlinkAus();                            // Cursor und Blinken aus
    ActualizeDisplay();                          // Werte ausgeben
    curpos = (byte)CURSORMARGINLEFT;             // Cursor auf Startposition
    lcd.setCursor(curpos,0);                     // auf Startposition setzen
  }
}

// Software - Display und Tasten -----------------------------------------------
uint8_t GetButtons(void)
{
  uint8_t rdbuttons  = (uint8_t)0;               // aktuelle Tasten
  uint8_t oldbuttons = (uint8_t)0;               // vorige Tasten

  rdbuttons = lcd.readButtons();                 // Tasten einlesen
  while (rdbuttons != 0)                         // Warten auf Taste losgelassen
  {
    oldbuttons = rdbuttons;                      // aktuellen Wert Uebernehmen
    rdbuttons  = lcd.readButtons();              // Tasten einlesen
  }
  return(oldbuttons);                            // Tastenwerte Uebergeben
}

void CheckButtons(uint8_t butt)
{
  byte zehner = 0;
  byte einer  = 0;
  
  if (butt != (uint8_t)0)                       // Ein Button wurde gedrueckt
  {
    actualtime = starttime = millis();          // Zeitmessung starten
    if (buttonpressed == false)
    {
      SaveTimeToFallback();                     // aktuelle zeit sichern
      buttonpressed = true;                     // Eine Taste wurde gedrueckt
    }
    CursorBlinkEin();                           // Cursor und Blinken ein
    if (butt & BUTTON_UP)                       // BUTTON UP gedrueckt
    {
      lcd.setCursor(curpos,0);                  // auf aktuelle Cursorposition setzen
      switch(curpos)                            // nach Position auswerten
      {
        // Tage max = 31
        case  0 : // Tag Zehner
          if (Get_Tag() > (byte)0)
          {
            zehner = GetZehner(Get_Tag());
          }
          if (zehner < (byte)3)
          {
            Set_Tag(Get_Tag() + (byte)10);
            if (Get_Tag() > (byte)31)
            {
              Set_Tag((byte)31);
            }
          }
          lcd.print(GetZehner(Get_Tag()), DEC);
          lcd.print(GetEiner(Get_Tag()) , DEC);
          DebugPrintValue("Tag Zehner = ", GetZehner(Get_Tag()));
          break;
        case  1 : // Tag Einer
          if (Get_Tag() > (byte)0)
          {
            zehner = GetZehner(Get_Tag());
            einer  = GetEiner(Get_Tag());
          }
          if (zehner < (byte)3)
          {
            if (einer < (byte)9)
            {
              Set_Tag(Get_Tag() + (byte)1);
            }
          }
          else
          {
            if ((byte)0 == einer)
            {
               Set_Tag(Get_Tag() + (byte)1);
            }
          }
          lcd.print(GetEiner(Get_Tag()),DEC);
          DebugPrintValue("Tag Einer = ",GetEiner(Get_Tag()));
          break;
        // Monat max = 12  
        case  3 : // Monat Zehner
          if (Get_Monat() > (byte)0)
          {
            zehner = GetZehner(Get_Monat());
          }
          if (zehner < (byte)1)
          {
            Set_Monat(Get_Monat() + (byte)10);
          }
          if (Get_Monat() > (byte)12)
          {
            Set_Monat((byte)12);
          }
          lcd.print(GetZehner(Get_Monat()), DEC);
          lcd.print(GetEiner(Get_Monat()) , DEC);
          DebugPrintValue("Monat Zehner = ", GetZehner(Get_Monat()));
          break;
        case  4 : // Monat Einer
          if (Get_Monat() > (byte)0)
          {
            zehner = GetZehner(Get_Monat());
            einer  = GetEiner(Get_Monat());
          }
          if (zehner < (byte)1)
          {
            if (einer < (byte)9)
            {
              Set_Monat(Get_Monat() + (byte)1);
            }
          }
          else
          {
            if (einer < (byte)2)
            {
               Set_Monat(Get_Monat() + (byte)1);
            }
          }
          lcd.print(GetEiner(Get_Monat()),DEC);
          DebugPrintValue("Monat Einer = ",GetEiner(Get_Monat()));
          break;
        // Jahr max = 99
        case  6 : // Jahr Zehner
          if (Get_Jahr() > (byte)0)
          {
            zehner = GetZehner(Get_Jahr());
          }
          if (zehner < (byte)9)
          {
            Set_Jahr(Get_Jahr() + 10);
          }
          lcd.print(GetZehner(Get_Jahr()),DEC);
          DebugPrintValue("Jahr Zehner = ",GetZehner(Get_Jahr()));
          break;
        case  7 : // Jahr Einer
          if (Get_Jahr() > (byte)0)
          {
            einer = GetEiner(Get_Jahr());
          }
          if (einer < (byte)9)
          {
            Set_Jahr(Get_Jahr() + (byte)1);
          }
          lcd.print(GetEiner(Get_Jahr()),DEC);
          DebugPrintValue("Jahr Einer = ",GetEiner(Get_Jahr()));
          break;
        // Stunde max = 23
        case  9 : // Stunde Zehner
          if (Get_Stunde() > (byte)0)
          {
            zehner = GetZehner(Get_Stunde());
          }
          if (zehner < (byte)2)
          {
            Set_Stunde(Get_Stunde() + (byte)10);
            if (Get_Stunde() > (byte)23)
            {
              Set_Stunde((byte)23);
            }
          }
          lcd.print(GetZehner(Get_Stunde()), DEC);
          lcd.print(GetEiner(Get_Stunde()) , DEC);
          DebugPrintValue("Stunde Zehner = ", GetZehner(Get_Stunde()));
          break;
        case 10 : // Stunde Einer
          if (Get_Stunde() > (byte)0)
          {
            zehner = GetZehner(Get_Stunde());
            einer  = GetEiner(Get_Stunde());
          }
          if (zehner < (byte)2)
          {
            if (einer < (byte)9)
            {
              Set_Stunde(Get_Stunde() + (byte)1);
            }
          }
          else
          {
            if (einer < (byte)3)
            {
               Set_Stunde(Get_Stunde() + (byte)1);
            }
          }
          lcd.print(GetEiner(Get_Stunde()),DEC);
          DebugPrintValue("Stunde Einer = ",GetEiner(Get_Stunde()));
          break;
        // Minute max = 59
        case 12 : // Minute Zehner
          if (Get_Minute() > (byte)0)
          {
            zehner = GetZehner(Get_Minute());
          }
          if (zehner < (byte)5)
          {
            Set_Minute(Get_Minute() + (byte)10);
          }
          lcd.print(GetZehner(Get_Minute()), DEC);
          lcd.print(GetEiner(Get_Minute()) , DEC);
          DebugPrintValue("Minute Zehner = ", GetZehner(Get_Minute()));
          break;
        case 13 : // Minute Einer
          if (Get_Minute() > (byte)0)
          {
            zehner = GetZehner(Get_Minute());
            einer  = GetEiner(Get_Minute());
          }
          if (einer < (byte)9)
          {
            Set_Minute(Get_Minute() + (byte)1);
          }
          lcd.print(GetEiner(Get_Minute()),DEC);
          DebugPrintValue("Minute Einer = ",GetEiner(Get_Minute()));
          break;
        default :
          break;
      }
      lcd.setCursor(curpos,0);                  // Cursor neu setzen
    }
    
    if (butt & BUTTON_DOWN)                     // BUTTON DOWN gedrueckt
    {
      lcd.setCursor(curpos,0);                  // auf aktuelle Cursorposition setzen
      switch(curpos)                            // nach Position auswerten
      {
        // Tage min = 1
        case  0 : // Tag Zehner
          if (Get_Tag() > (byte)0)
          {
            zehner = GetZehner(Get_Tag());
          }
          if (zehner > (byte)0)
          {
            Set_Tag(Get_Tag() - (byte)10);
          }
          if (Get_Tag() < (byte)1)
          {
            Set_Tag((byte)1);
          }
          lcd.print(GetZehner(Get_Tag()), DEC);
          lcd.print(GetEiner(Get_Tag()) , DEC);
          DebugPrintValue("Tag Zehner = ", GetZehner(Get_Tag()));
          break;
        case  1 : // Tag Einer
          if (Get_Tag() > (byte)0)
          {
            einer = GetEiner(Get_Tag());
          }
          if (einer > (byte)0)
          {
              Set_Tag(Get_Tag() - (byte)1);
          }
          if ((byte)0 == Get_Tag())
          {
            Set_Tag((byte)1);
          }
          lcd.print(GetEiner(Get_Tag()), DEC);
          DebugPrintValue("Tag Einer = ", GetEiner(Get_Tag()));
          break;
        // Monat min = 1
        case  3 : // Monat Zehner
          if (Get_Monat() > (byte)0)
          {
            zehner = GetZehner(Get_Monat());
          }
          if (zehner > (byte)0)
          {
            Set_Monat(Get_Monat() - (byte)10);
          }
          if (Get_Monat() < (byte)1)
          {
            Set_Monat((byte)1);
          }
          lcd.print(GetZehner(Get_Monat()), DEC);
          lcd.print(GetEiner(Get_Monat()) , DEC);
          DebugPrintValue("Monat Zehner = ", GetZehner(Get_Monat()));
          break;
        case  4 : // Monat Einer
          if (Get_Monat() > (byte)0)
          {
            einer = GetEiner(Get_Monat());
          }
          if (einer > (byte)0)
          {
              Set_Monat(Get_Monat() - (byte)1);
          }
          if ((byte)0 == Get_Monat())
          {
            Set_Monat((byte)1);
          }
          lcd.print(GetEiner(Get_Monat()), DEC);
          DebugPrintValue("Monat Einer = ", GetEiner(Get_Monat()));
          break;
        // Jahr min = 0
        case  6 : // Jahr Zehner
          if (Get_Jahr() > (byte)0)
          {
            zehner = GetZehner(Get_Jahr());
          }
          if (zehner > (byte)0)
          {
            Set_Jahr(Get_Jahr() - (byte)10);
          }
          lcd.print(GetZehner(Get_Jahr()),DEC);
          DebugPrintValue("Jahr Zehner = ",GetZehner(Get_Jahr()));
          break;
        case  7 : // Jahr Einer
          if (Get_Jahr() > (byte)0)
          {
            einer = GetEiner(Get_Jahr());
          }
          if (einer > (byte)0)
          {
            Set_Jahr(Get_Jahr() - (byte)1);
          }
          lcd.print(GetEiner(Get_Jahr()),DEC);
          DebugPrintValue("Jahr Einer = ",GetEiner(Get_Jahr()));
          break;
        // Stunde min = 0
        case  9 : // Stunde Zehner
          if (Get_Stunde() > (byte)0)
          {
            zehner = GetZehner(Get_Stunde());
          }
          if (zehner > (byte)0)
          {
            Set_Stunde(Get_Stunde() - (byte)10);
          }
          lcd.print(GetZehner(Get_Stunde()), DEC);
          DebugPrintValue("Stunde Zehner = ", GetZehner(Get_Stunde()));
          break;
        case 10 : // Stunde Einer
          if (Get_Stunde() > (byte)0)
          {
            zehner = GetZehner(Get_Stunde());
            einer  = GetEiner(Get_Stunde());
          }
          if (einer > (byte)0)
          {
            Set_Stunde(Get_Stunde() - (byte)1);
          }
          lcd.print(GetEiner(Get_Stunde()) ,DEC);
          DebugPrintValue("Stunde Einer = ",GetEiner(Get_Stunde()));
          break;
        // Minute min = 0
        case 12 : // Minute Zehner
          if (Get_Minute() > (byte)0)
          {
            zehner = GetZehner(Get_Minute());
          }
          if (zehner > (byte)0)
          {
            Set_Minute(Get_Minute() - (byte)10);
          }
          lcd.print(GetZehner(Get_Minute()), DEC);
          DebugPrintValue("Minute Zehner = ", GetZehner(Get_Minute()));
          break;
        case 13 : // Minute Einer
          if (Get_Minute() > (byte)0)
          {
            einer = GetEiner(Get_Minute());
          }
          if (einer > (byte)0)
          {
            Set_Minute(Get_Minute() - (byte)1);
          }
          lcd.print(GetEiner(Get_Minute()),DEC);
          DebugPrintValue("Minute Einer = ",GetEiner(Get_Minute()));
          break;
        default :
          break;
      }
      lcd.setCursor(curpos,0);                  // Cursor neu setzen
    }
    
    if (butt & BUTTON_LEFT) 
    {
      if (curpos > (byte)CURSORMARGINLEFT)
      {
        curpos--;
      }
      if ((curpos == 2) | (curpos == 5) | (curpos == 8) | (curpos == 11)) // '.',':',' ' ueberspringen
      {
        curpos--;
      }
      lcd.setCursor(curpos,0);                  // Cursor neu setzen
    }    
    
    if (butt & BUTTON_RIGHT) 
    {
      if (curpos < (byte)CURSORMARGINRIGHT)
      {
        curpos++;
      }
      if ((curpos == 2) | (curpos == 5) | (curpos == 8) | (curpos == 11)) // '.',':',' ' ueberspringen
      {
        curpos++;
      }
      lcd.setCursor(curpos,0);                  // Cursor neu setzen
    }
    
    DebugPrintValue("Button gedrueckt - Cursor ein - pos = ",curpos);

    if (butt & BUTTON_SELECT)                  // BUTTON SELECT gedrueckt
    {
      CursorBlinkAus();                        // Cursor und Blinken aus
      DebugPrint("Select gedrueckt und Cursor aus");
      buttonpressed       = false;             // irgendeine Taste gedrueckt
      selectbuttonpressed = true;              // Select Taste gedrueckt
    }
  }
}

void ActualizeDisplay(void)
{
  DisplayTime(Get_Tag(),Get_Monat(),Get_Jahr(),Get_Stunde(),Get_Minute());      // Zeitzeile ausgeben
  DisplayMesswert(Get_Temperatur(), Get_Luftfeuchte());// Messzeile ausgeben
}

void DisplayTime(byte t, byte m, byte j, byte s, byte mn)
{
  lcd.setCursor(0, 0); // obere Zeile
  // Tag
  lcd.print(GetZehner(t),DEC);
  lcd.print(GetEiner(t) ,DEC);
  lcd.print(".");
  // Monat
  lcd.print(GetZehner(m),DEC);
  lcd.print(GetEiner(m) ,DEC);
  lcd.print(".");
  // Jahr
  lcd.print(GetZehner(j),DEC);
  lcd.print(GetEiner(j) ,DEC);
  // Trenn
  lcd.print(" ");
  // Stunde
  lcd.print(GetZehner(s),DEC);
  lcd.print(GetEiner(s) ,DEC);
  lcd.print(":");
  // Minute
  lcd.print(GetZehner(mn),DEC);
  lcd.print(GetEiner(mn) ,DEC);
  
}

void DisplayMesswert(byte t, byte f)
{
  char grad[2] = {(char)223,(char)0}; // Gradzeichen nach Datenblatt LCD 44780
  
  lcd.setCursor(0, 1); // untere Zeile
  // Luftfeuchte
  lcd.print(GetZehner(f));
  lcd.print(GetEiner(f));
  lcd.print("%");
  // Trenn
  lcd.print(" ");
  // Temperatur
  lcd.print(GetZehner(t));
  lcd.print(GetEiner(t));
  lcd.print(grad);
}

byte GetZehner(byte x)
{
  return(x / 10);
}

byte GetEiner(byte y)
{
  return(y - (GetZehner(y) * 10));
}

void CheckTimeOverrun(void)
{
  if (actualtime < starttime)                      // Ueberlauf von millis() oder Programmfehler
  {
    starttime = actualtime = millis();
    DebugPrint("Ueberlauf millis()");
  }
}

void CursorBlinkAus(void)
{
  lcd.noBlink();
  lcd.noCursor();
  DebugPrint("Cursor ist aus");
}

void CursorBlinkEin(void)
{
  lcd.cursor();
  lcd.blink();  
  DebugPrint("Cursor ist an");
}

